{% if this.content %}
<div class="center">
  {{ this.content }}
</div>
{% endif %}
{% if this.button_type and this.button_content %}
<div class="row">
  {% if this.button_type == "text" %}
  <a href="{{ this.button_link }}"  class="content-button text-button" target="_blank" rel="noopener noreferrer">{{ this.button_content | safe }}</a>
  {% elif this.button_type == "image" %}
  <a href="{{ this.button_link }}" class="center content-button image-button" target="_blank" rel="noopener noreferrer">
    <img alt="Action Button" src="{{ this.button_content | url }}" class="center">
  </a>
  {% endif %}
</div>
{% endif %}
