#!/usr/bin/env python
"""A package for generating ctypes wrappers from a C header file.

See the docstring for wrap.py for usage.

The files 'lex.py' and 'yacc.py' are from PLY (http://www.dabeaz.com/ply),
which was written by David M. Beazley but have been modified slightly
for this tool.
"""